@@ -107,6 +107,38 @@ struct prior {
 	, num_sites( prior_num_sites )
 	{
 	}
+
+	/**
+	 * Calculate the logarithm of the prior using the given g.
+	 *
+	 * The prior is lambda(1-g) / [1 - lambda(1-g)]
+	 *
+	 * Deals with the case that g==0 specially
+	 */
+	inline
+	double
+	calculate_log_prior( double g ) const {
+        BOOST_ASSERT( 0 <= g );
+        BOOST_ASSERT( g <= 1 );
+	    if( g > 0. ) {
+            const double prior_term = lambda.get_lambda() * ( 1. - g );
+            return std::log( prior_term / ( 1. - prior_term ) );
+	    } else {
+	        return lambda.get_log_lambda_ratio();
+	    }
+	}
+
+    /**
+     * Calculate the the prior using the given g.
+     */
+    inline
+    double
+    calculate_prior( double g ) const {
+        BOOST_ASSERT( 0 <= g );
+        BOOST_ASSERT( g <= 1 );
+        const double term = lambda.get_lambda() * ( 1. - g );
+        return term / ( 1. - term );
+    }
 };
 
 
@@ -175,23 +207,36 @@ public:
 	}
 
 
-	/** Calculate Z for the W-mer at the given position. Can over-ride the prior with optional argument. */
+	/** Calculate Z for the W-mer at the given position. Can over-ride the position-specific prior with optional argument. */
 	double
-	calculate_Z( global_pos_t global_pos, bool rev_comp, boost::optional< double > g = boost::optional< double >() ) {
+	calculate_Z(
+        global_pos_t global_pos,
+        bool rev_comp,
+        boost::optional< double > g = boost::optional< double >()
+    ) {
+	    // The position
         const int seq = seqan::getSeqNo( global_pos, _data.get_string_set_limits() );
         const int offset = seqan::getSeqOffset( global_pos, _data.get_string_set_limits() );
+
+        // Likelihood of binding under model
 	    const double log_p_binding =
             rev_comp
                 ? w_mer_binding_log_likelihood< true  >( bs, _data.get_W_mer( W(), global_pos ) )
                 : w_mer_binding_log_likelihood< false >( bs, _data.get_W_mer( W(), global_pos ) )
                 ;
+
+	    // Likelihood under background
 	    const double log_p_background = bg.wmer_log_likelihood( seq, offset );
-	    const double lambda = _prior.lambda.get_lambda();
-	    const double odds_ratio = std::exp( log_p_background - log_p_binding );
+
+	    // Log Bayes factor
+	    const double log_bayes_factor = log_p_binding - log_p_background;
+
+	    // Prior
 	    const double g_n = g ? *g : _prior._psp->get_g( seq, offset );
-	    const double prior_term = ( ( 1. + g_n / ( 1. - g_n ) ) / lambda - 1. );
-	    const double Z = 1. / ( 1. + prior_term * odds_ratio );
-	    return Z;
+	    const double log_prior = _prior.calculate_log_prior( g_n );
+
+	    // Calculate Z
+	    return lr_to_prob( std::exp( log_prior + log_bayes_factor ) );
 	}
 
 
